#include <stdlib.h> 
#include <stdio.h> 
/*
  (c) Nikolai Yu. Zolotykh, 2004
*/

/**
  Horowitz-Sahni algorithm (1974)
  for 0-1 knapsack problem

  max z = p[0]*x[0] + p[1]*x[1] + ... + p[n]*x[n]
  subject to
          w[0]*x[0] + w[1]*x[1] + ... + w[n]*x[n] <= c

          x[j] = {0,1} (j=1,2,...,n)
          
          w[j] > 0, 
          p[j] > 0


Example:
  n = 7;
  c = 50;

  p[0] = 70; p[1] = 20; p[2] = 39; p[3] = 37; p[4] = 7; p[5] = 5; p[6] = 10;
  w[0] = 31; w[1] = 10; w[2] = 20; w[3] = 19; w[4] = 4; w[5] = 3; w[6] = 6;

Solution: z=7
  x[0] = 1; x[1] = 0; x[2] = 0; x[3] = 1; x[4] = 0; x[5] = 0; x[6] = 0;
*/
void knapsack_hs(int n, long* w, long c, long* p, long* z, long* x);


/**
  Sorting indices.

  Sorting indices such that 
     p[0]/w[0] >= p[1]/w[1] >= ... >= p[n]/w[n] 
  The correspondent permutation stores on the array perm.
*/
void knapsack_sort(long* a, long* b, int n, int* perm);

/**
  Restore the initial permutation of variables.
*/
void knapsack_restore(long* x, int n, int* perm);


/*
  This is sorting function used in knapsack procedures
  Here I use my template quick_sort.c
  
  (c) Nikolai Yu. Zolotykh, 2004
*/

#include <stdlib.h>

#define QUICK_SORT_SWAP(i,j,temp,temp2) temp=a[i]; a[i]=a[j]; a[j]=temp; temp=b[i]; b[i]=b[j]; b[j]=temp; temp2=perm[i]; perm[i]=perm[j]; perm[j]=temp2; 
#define QUICK_SORT_LESS_EQUAL(i,j) ((float)a[i]/(float)b[i] >= (float)a[j]/(float)b[j] ? 1 : 0)

void quick_sort(long* a, long* b, int n, int* perm)
{
  long temp;
  int pivot_ind;
  int j;
  int temp2;
 
  if (n <= 1) return;
  
  QUICK_SORT_SWAP(n/2, 0, temp, temp2)
  pivot_ind = 0;
  
  for (j = 1 ; j < n; j++) 
  {
    if (QUICK_SORT_LESS_EQUAL(j, 0)) 
    {
      pivot_ind++;
      QUICK_SORT_SWAP(j, pivot_ind, temp, temp2)
    }
  }
  QUICK_SORT_SWAP(pivot_ind, 0, temp, temp2)

  quick_sort(a, b, pivot_ind, perm);
  quick_sort(&a[pivot_ind + 1], &b[pivot_ind + 1], n - pivot_ind - 1, &perm[pivot_ind + 1]);
}

void knapsack_sort(long* a, long* b, int n, int* perm)
{
  int j;

  for (j = 0; j < n; j++)
    perm[j] = j;

  quick_sort(a, b, n, perm);
}

void knapsack_restore(long* x, int n, int* perm)
{
  int j;

  long* x_tmp;
  x_tmp = (long*) malloc(n * sizeof(long));

  for (j = 0; j < n; j++)
    x_tmp[perm[j]] = x[j];

  for (j = 0; j < n; j++)
    x[j] = x_tmp[j];

  free(x_tmp);
}

void knapsack_hs(int n, long* w, long c, long* p, long* z, long* x)
/*
  Horowitz-Sahni algorithm (1974)
  for 0-1 knapsack problem
*/

{
  long z_cur = 0;
  long c_cur = c;
  int j = 0;

  long s;
  int r;
  long u;
  int k;
  int i;

  long* x_cur;

  int flag;

  (*z) = 0;
  p[n] = 0;
  w[n] = c + 1; //INF;

  x_cur = (long *) malloc((n + 1) * sizeof(long));
  
  while(1)
  {
    // r = min{i: w[j] + w[j+1] + ... + w[i] > c_cur}
    for(s = 0, r = j; s <= c_cur; r++)
      s += w[r];
    r--;

    for(u = 0, k = j; k < r; k++)
      u += p[k];

    u += (c_cur - s + w[r]) * p[r] / w[r];

    if (z_cur + u > (*z))
    {
      do
      {
         while (w[j] <= c_cur)
         {
            c_cur -= w[j];
            z_cur += p[j];
            x_cur[j] = 1;
            j++;
         }
         if (j < n)
         {
           x_cur[j] = 0;
           j++;
         }
      }
      while (j == n-1);

      if (j < n - 1) continue;

      if (z_cur > (*z))
      {
        (*z) = z_cur;
        for (k = 0; k < n; k++)
           x[k] = x_cur[k];
      }
      j = n - 1;
      if (x_cur[n - 1] == 1)
      {
        c_cur += w[n - 1];
        z_cur -= p[n - 1];
        x[n - 1] = 0;
      }
    }

    for (flag = 0, i = j - 1; i >= 0; i--)
    {
      if (x_cur[i] == 1) 
      {
        flag = 1;
        break;
      }
    }

    if (!flag) break;

    c_cur += w[i];
    z_cur -= p[i];
    x_cur[i] = 0;
    
    j = i + 1;

  }
  
  free(x_cur);
}


void main()
{
  long* p;
  long* w;
  long* x;
  int j;
  long c;
  int n;
  long z;
  int* perm;
  
  printf("n = ? ");
  scanf("%d", &n);
  printf("OK\n");

  p = (long *) malloc((n+1) * sizeof(long));
  w = (long *) malloc((n+1) * sizeof(long));
  x = (long *) malloc((n+1) * sizeof(long));
  perm = (int*) malloc(n * sizeof(int));
  
  printf("c = ? ");
  scanf("%d", &c);
  printf("OK\n");

  printf("p[j] = ? ");
  for (j = 0; j < n; j++)
    scanf("%d", &p[j]);
  printf("OK\n");

  printf("w[j] = ? ");
  for (j = 0; j < n; j++)
    scanf("%d", &w[j]);
  printf("OK\n");
  
  knapsack_sort(p, w, n, perm);
  knapsack_hs(n, w, c, p, &z, x);
  knapsack_restore(x, n, perm);

  printf("z = %d\n", z);

  for (j = 0; j < n; j++)
    printf("%5i ", x[j]);
  printf("\n");

  free(p);
  free(w);
  free(x);
}
